#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Command Line Interface - Utilities."""

from collections.abc import Iterator
from logging import FATAL, getLogger
from pathlib import Path

import click

from ucdp.finder import find
from ucdp.pathutil import improved_glob

PathType = click.Path(path_type=Path)


def auto_top(ctx, param, incomplete):
    """Autocompletion for TOP."""
    getLogger().setLevel(FATAL)
    infos = find(patterns=[f"{incomplete}*"], glob=True)
    return [str(info.topmodref) for info in infos]


def auto_path(ctr, param, incomplete):
    """Autocompletion for Paths."""
    return [str(path) for path in improved_glob(Path(f"{incomplete}*"))]


arg_top = click.argument("top", envvar="UCDP_TOP", shell_complete=auto_top)
arg_tops = click.argument("tops", nargs=-1, envvar="UCDP_TOP", shell_complete=auto_top)
opt_topsfile = click.option(
    "--tops-file",
    type=click.Path(path_type=Path),
    default=[],
    multiple=True,
    help="File with Top Module References",
)

opt_path = click.option(
    "--path",
    "-p",
    default=[],
    multiple=True,
    envvar="UCDP_PATH",
    shell_complete=auto_path,
    help="""
Search Path For Data Model And Template Files.
This option can be specified multiple times.
Environment Variable 'UCDP_PATH'.
""",
)
opt_filelist = click.option(
    "--filelist",
    "-f",
    multiple=True,
    help="Filelist Names. Environment Variable 'UCDP_FILELIST'.",
    envvar="UCDP_FILELIST",
)
opt_target = click.option(
    "--target",
    "-t",
    help="Filter File List for Target. Environment Variable 'UCDP_TARGET'.",
    envvar="UCDP_TARGET",
)
opt_show_diff = click.option(
    "--show-diff",
    "-s",
    default=False,
    is_flag=True,
    help="Show What Changed. Environment Variable 'UCDP_SHOW_DIFF'.",
    envvar="UCDP_SHOW_DIFF",
)
opt_maxlevel = click.option("--maxlevel", "-L", type=int, help="Limit to maximum number of hierarchy levels.")
opt_dry_run = click.option("--dry-run", default=False, is_flag=True, help="Do nothing.")
opt_maxworkers = click.option(
    "--maxworkers",
    "-J",
    type=int,
    help="Maximum Number of Processes.",
    envvar="UCDP_MAXWORKERS",
)
opt_defines = click.option(
    "--define",
    "-D",
    multiple=True,
    type=str,
    help="Defines set on the datamodel. Environment Variable 'UCDP_DEFINES'",
    envvar="UCDP_DEFINES",
)
opt_file = click.option(
    "--file",
    "-o",
    type=click.File("w"),
    shell_complete=auto_path,
    help="Output to file instead of STDOUT",
)
opt_filepath = click.option(
    "--file",
    "-o",
    type=click.Path(path_type=Path),
    shell_complete=auto_path,
    help="Output to file instead of STDOUT",
)
opt_tag = click.option(
    "--tag",
    "-G",
    default=[],
    multiple=True,
    help="Filter Modules by Tag Name or Wildcard.",
)
arg_template_filepaths = click.argument(
    "template_filepaths",
    type=PathType,
    shell_complete=auto_path,
    nargs=-1,
    envvar="UCDP_TEMPLATE_FILEPATHS",
)
opt_local = click.option(
    "--local/--no-local",
    "-l/-L",
    default=None,
    is_flag=True,
    help="List local/non-local modules only.",
)
opt_check = click.option(
    "--check",
    default=False,
    is_flag=True,
    help="Report an error if any file changes.",
)
opt_create = click.option(
    "--create",
    "-c",
    default=False,
    is_flag=True,
    help="Create missing inplace files.",
)


def defines2data(defines: list[str]) -> dict[str, str]:
    """Convert defines to data."""
    return dict(define.split("=", 1) if "=" in define else (define, None) for define in defines)


def read_file(filepath: Path) -> Iterator[str]:
    """Read File."""
    with filepath.open() as file:
        for line in file:
            line = line.strip()  # noqa: PLW2901
            if line.startswith("#"):
                continue
            if line:
                yield line



#
# MIT License
#
# Copyright (c) 2024-2025 nbiotcloud
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Top Reference."""

import re
from pathlib import Path
from typing import ClassVar, Union

from .modref import ModRef
from .object import Field, LightObject, PosArgs

RE_TOPMODREF = re.compile(
    # [tb]#
    r"((?P<tb>[a-zA-Z_0-9_\.]+)#)?"
    # top
    r"(?P<top>[a-zA-Z_0-9_\.]+)"
    # [-sub]
    r"(-(?P<sub>[a-zA-Z_0-9_\.]+))?"
)
PAT_TOPMODREF = "[tb_lib.tb#]top_lib.top[-sub_lib.sub]"
RE_MODREF = r"[a-zA-Z][a-zA-Z_0-9]*\.[a-zA-Z][a-zA-Z_0-9]*"


class TopModRef(LightObject):
    """
    Top Module Reference.

    Args:
        top: Top Module Reference

    Attributes:
        sub: Sub Module Reference
        tb: Testbench Module Reference

    ??? Example "Top Reference Examples"
        Example:

            >>> import ucdp as u
            >>> u.TopModRef.cast('top_lib.top_mod')
            TopModRef(ModRef('top_lib', 'top_mod'))
            >>> u.TopModRef.cast('top_lib.top_mod-sub_lib.sub_mod')
            TopModRef(ModRef('top_lib', 'top_mod'), sub='sub_lib.sub_mod')
            >>> u.TopModRef.cast('mod_tb_lib.mod_tb#top_lib.top_mod')
            TopModRef(ModRef('top_lib', 'top_mod'), tb=ModRef('mod_tb_lib', 'mod_tb'))
            >>> u.TopModRef.cast(TopModRef(ModRef('top_lib', 'top_mod')))
            TopModRef(ModRef('top_lib', 'top_mod'))

            Invalid Pattern:

            >>> TopModRef.cast('lib.mod:c-ls.1')
            Traceback (most recent call last):
            ..
            ValueError: 'lib.mod:c-ls.1' does not match pattern '[tb_lib.tb#]top_lib.top[-sub_lib.sub]'
    """

    top: ModRef
    sub: str | None = Field(default=None, pattern=RE_MODREF)
    tb: ModRef | None = None

    _posargs: ClassVar[PosArgs] = ("top",)

    def __init__(self, top: ModRef, sub: str | None = None, tb: ModRef | None = None):
        super().__init__(top=top, sub=sub, tb=tb)  # type: ignore[call-arg]

    def __str__(self) -> str:
        result = str(self.top)
        if self.sub:
            result = f"{result}-{self.sub}"
        if self.tb:
            result = f"{self.tb}#{result}"
        return result

    @staticmethod
    def cast(value: Union["TopModRef", Path, str]) -> "TopModRef":
        """Cast `value` to `TopModRef`."""
        if isinstance(value, TopModRef):
            return value

        if isinstance(value, Path):
            path = value.resolve()
            modname = path.stem
            libname = path.parent.name
            return TopModRef(top=ModRef(libname=libname, modname=modname))

        mat = RE_TOPMODREF.fullmatch(value)
        if mat:
            top = ModRef.cast(mat.group("top"))
            sub = mat.group("sub")
            tb = ModRef.cast(mat.group("tb")) if mat.group("tb") else None
            return TopModRef(top=top, sub=sub, tb=tb)

        raise ValueError(f"{value!r} does not match pattern {PAT_TOPMODREF!r}")
